package Glib::Object::Introspection::Install::Files;

$self = {
          'inc' => '-I/usr/local/include/gobject-introspection-1.0 -I/usr/local/include -I/usr/local/include/glib-2.0 -I/usr/local/lib/glib-2.0/include -I/usr/local/include -pthread -I/usr/local/include/glib-2.0 -I/usr/local/lib/glib-2.0/include -I/usr/local/include',
          'typemaps' => [],
          'deps' => [
                      'Glib'
                    ],
          'libs' => '-L/usr/local/lib -lgirepository-1.0 -lgobject-2.0 -lglib-2.0 -lintl -Wl,--export-dynamic -L/usr/local/lib -pthread -lgmodule-2.0 -lglib-2.0 -lintl -L/usr/local/lib -lffi'
        };


# this is for backwards compatiblity
@deps = @{ $self->{deps} };
@typemaps = @{ $self->{typemaps} };
$libs = $self->{libs};
$inc = $self->{inc};

	$CORE = undef;
	foreach (@INC) {
		if ( -f $_ . "/Glib/Object/Introspection/Install/Files.pm") {
			$CORE = $_ . "/Glib/Object/Introspection/Install/";
			last;
		}
	}

1;
